
/*
	HSVAdjust plugin for AviSynth -- Let's you adjust the hue, saturation and value, lightness or intensity.

	Copyright (C) 2011 Wilbert Dijkhof
  
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.
	 
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
		
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	
	The author can be contacted at:
	Wilbert Dijkhof
	wilbertdijkhof at yahoo.com
*/

#include "HSVAdjust.h"


/**********************
*****  HSVAdjust  *****
**********************/

HSVAdjust::HSVAdjust( PClip _child, double _hue, double _sat, double _bright, double _cont, const char* _brightness,
                            IScriptEnvironment* env ) 
  : GenericVideoFilter(_child)
{
    if (!vi.IsRGB())
      env->ThrowError("HSVAdjust requires RGB input");

	brightness = value;
	if (_brightness) {
		if (!lstrcmpi(_brightness, "value"))
			brightness = value;
		else if (!lstrcmpi(_brightness, "lightness"))
			brightness = lightness;
		else
			env->ThrowError("HSVAdjust: invalid \"brightness\" parameter (must be brightness=\"value\", or \"lightness\")");
	}

    Sat = (int) (_sat * 512);
    Cont = (int) (_cont * 512);
    Bright = (int) (_bright * 512);
    Hue = (int)(_hue);

    int j, k, h, Hi, f;
	double div;
    if (vi.IsRGB24()) {
      channels = 3;
	} else { // RGB32
	  channels = 4;
	}
    for (j=-256; j<256; j++) {
      mapH[256+j][0] = (_hue > 0) ? int(_hue + 0.5) : int(_hue - 0.5);
      for (k=1; k<256; k++) {
		div = 60.0*j/k + _hue;
        mapH[256+j][k] = (div > 0) ? int(div + 0.5) : int(div - 0.5);
	  }
	}
    for (h=0; h<359; h++) {
	  Hi = int(h/60);
	  f = int(512*(h/60.0 - Hi)); // 0<=f<512
	  mapfHi[h] = int(Hi<<9|f);
	}
    for (j=0; j<256; j++) {
	  mapS[j][0] = 0;
      for (k=1; k<256; k++) {
        mapS[j][k] = int(double(j)/double(k) * Sat + 0.5);
		mapS[j][k] = min(255*Sat,max(0,mapS[j][k]));
	  }
	}
}


PVideoFrame __stdcall HSVAdjust::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame src = child->GetFrame(n, env);
	env->MakeWritable(&src);

	BYTE* srcp = src->GetWritePtr();

	int src_pitch = src->GetPitch();
	int height = src->GetHeight();
	int row_size = src->GetRowSize();

	int x,y;

	int r, g, b, M, m, Chroma, Ha, f, Hi, fHi, Chromaa, Va, Sa, ma;
	for (y=0; y<height; y++) {
        for (x=0; x<row_size; x+=channels) {
			/* brightness and contrast */
			/* hue and saturation */
			b = srcp[x];
			g = srcp[x+1];
			r = srcp[x+2];
			m = min(min(b,g),r);
			M = max(max(b,g),r);
			Chroma = M - m;
			if (r==M) {
				Ha = mapH[256+g-b][Chroma];
			} else if (g==M) {
				Ha = mapH[256+b-r][Chroma] + 120;
			} else { // b==M
				Ha = mapH[256+r-g][Chroma] + 240;
			}

			// wrap hue value in [0,360[
			while (Ha>=360) {
				Ha = Ha-360;
			}
			while (Ha<0) {
				Ha = Ha+360;
			}

			fHi = mapfHi[Ha];
			f = fHi&0x1ff;
			Hi = (BYTE)(fHi>>9);

			if (brightness == value) {
				Sa = mapS[Chroma][M]; // S'=C/V
				Va = Bright + Cont*M; // V'=vb+vc*V
				Va = min(255*512,max(0,Va));
				Chromaa = ScaledValue(Va*Sa); // C'=V'*S'
				ma = Va - Chromaa;
			} else { // lightness; L=(M+m)/2
				if (M+m<=255) { // S'=C/2L for 2L<=255
					Sa = mapS[Chroma][M+m];
					Va = Bright + (Cont*(M+m)>>1); // L'=vb+vc*L
					Va = min(255*512,max(0,Va));
					Chromaa = ScaledValue(2*Va*Sa); // C'=2L'*S'
				} else { // S' = C/(2*(255-L)) for 2L>255
					Sa = mapS[Chroma][510-(M+m)];
					Va = Bright + (Cont*(M+m)>>1); // L'=vb+vc*L
					Va = min(255*512,max(0,Va));
					Chromaa = ScaledValue(((130560-Va)*Sa)<<1); // C'=2*(255-L')*S'; V'=L'
				}
				ma = Va - (Chromaa>>1); // m'=L'-C'/2
			}

            /*
            S' = s*S, H' = h+H, V' = vb+vc*V, m' = V'-C'
            
            Hi	(R-m')/C'	(G-m')/C'	(B-m')/C'
            0		1			f			0
            1		1-f			1			0
            2		0			1			f
            3		0			1-f			1
            4		f			0			1
            5		1			0			1-f
            */

			switch (Hi) {
			case 0:
				b = ScaledValue(ma);
				g = ScaledValue(ScaledValue(f*Chromaa)+ma);
				r = ScaledValue(Chromaa+ma);
				break;
			case 1:
				b = ScaledValue(ma);
				g = ScaledValue(Chromaa+ma);
				r = ScaledValue(ScaledValue((512-f)*Chromaa)+ma);
				break;
			case 2:
				b = ScaledValue(ScaledValue(f*Chromaa)+ma);
				g = ScaledValue(Chromaa+ma);
				r = ScaledValue(ma);
				break;
			case 3:
				b = ScaledValue(Chromaa+ma);
				g = ScaledValue(ScaledValue((512-f)*Chromaa)+ma);
				r = ScaledValue(ma);
				break;
			case 4:
				b = ScaledValue(Chromaa+ma);
				g = ScaledValue(ma);
				r = ScaledValue(ScaledValue(f*Chromaa)+ma);
				break;
			case 5:
				b = ScaledValue(ScaledValue((512-f)*Chromaa)+ma);
				g = ScaledValue(ma);
				r = ScaledValue(Chromaa+ma);
				break;
			}
			srcp[x] = min(255,max(0,b));
			srcp[x+1] = min(255,max(0,g));
			srcp[x+2] = min(255,max(0,r));
		}
		srcp += src_pitch;
	}

	return src;
}

AVSValue __cdecl HSVAdjust::Create(AVSValue args, void* user_data, IScriptEnvironment* env)
{
    return new HSVAdjust(args[0].AsClip(),
    					 args[1].AsFloat(0.0),		// hue
    					 args[2].AsFloat(1.0),		// sat
	    				 args[3].AsFloat(0.0),		// bright
		    			 args[4].AsFloat(1.0),		// cont
			    		 args[5].AsString(0),		// brightness
    					 env);
}

AVSValue __cdecl HSVAdjust::CreateHSL(AVSValue args, void*, IScriptEnvironment* env)
{
  return new HSVAdjust(args[0].AsClip(), args[1].AsFloat(0.0), args[2].AsFloat(1.0), args[3].AsFloat(0.0), args[4].AsFloat(1.0), args[5].AsString("lightness"), env);
}



/**********************
*****  HSIAdjust  *****
**********************/

HSIAdjust::HSIAdjust( PClip _child, double _hue, double _sat, double _bright, double _cont, IScriptEnvironment* env ) 
  : GenericVideoFilter(_child)
{
    if (!vi.IsRGB())
      env->ThrowError("HSIAdjust requires RGB input");

    Sat = int(_sat * 512);
    Cont = int(_cont * 512);
    Bright = int(_bright * 512);
    Hue = int(_hue);
    coef = int(1024/3.0+0.5);

    int j, k, h, Hi, f;
	double div;

    if (vi.IsRGB24()) {
      channels = 3;
	} else { // RGB32
	  channels = 4;
	}

    for (j=0; j<256; j++) {
      mapH[j][0] = (_hue > 0) ? int(_hue + 0.5) : int(_hue - 0.5);
      for (k=1; k<256; k++) {
        div = 60.0*j/k + _hue;
        mapH[j][k] = (div > 0) ? int(div + 0.5) : int(div - 0.5);
	  }
	}

    for (h=0; h<359; h++) {
	Hi = int(h/60);
	f = int(512*(h/60.0 - Hi)); // 0<=f<512
	mapfHi[h] = int(Hi<<9|f);
	}
    for (j=0; j<1024; j++) {
      mapI[j] = int(j/3.0 + 0.5);
	}
}


PVideoFrame __stdcall HSIAdjust::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame src = child->GetFrame(n, env);
	env->MakeWritable(&src);

	BYTE* srcp = src->GetWritePtr();

	int src_pitch = src->GetPitch();
	int height = src->GetHeight();
	int row_size = src->GetRowSize();

	int x,y;

	int r, g, b, M, m, med, SumRGB, S, lambda, sign, Ha, Ia, Sa, fHi, f, Hi;

	for (y=0; y<height; y++) {
        for (x=0; x<row_size; x+=channels) {
			/* brightness and contrast */
			/* hue and saturation */
			b = srcp[x];
			g = srcp[x+1];
			r = srcp[x+2];
			M = max(max(b,g),r);
			m = min(min(b,g),r);
			SumRGB = r+g+b;
			med = SumRGB-m-M;

			if (SumRGB>=3*med) {
				S = (3*M-SumRGB)>>1;
			} else {
				S = (SumRGB-3*m)>>1;
			}

			if ((r>g)&&(g>=b)) {
				lambda = 0;
				sign = 1;
			} else if ((g>=r)&&(r>b)) {
				lambda = 1;
				sign = -1;
			} else if ((g>b)&&(b>=r)) {
				lambda = 2;
				sign = 1;
			} else if ((b>=g)&&(g>r)) {
				lambda = 3;
				sign = -1;
			} else if ((b>r)&&(r>=g)) {
				lambda = 4;
				sign = 1;
			} else if ((r>=b)&&(b>g)) {
				lambda = 5;
				sign = -1;
			} else { // r=g=b
				lambda = 0;
				sign = 1;
			}

			Ha = mapH[(S-sign*(SumRGB-3*med))>>1][S];

			Ha += 60*lambda;

			while (Ha>=360) {
				Ha = Ha-360;
			}
			while (Ha<0) {
				Ha = Ha+360;
			}

			Sa = S*Sat;
			Sa = min(255*512,max(0,Sa));
			Ia = Bright + ((coef*Cont*SumRGB+512)>>10); // I'=vb+vc*I
			Ia = min(255*512,max(0,Ia));

			fHi = mapfHi[Ha];
			f = fHi&0x1ff;
			Hi = (BYTE)(fHi>>9); // Hi=lambda for hue=0

            /*
            S' = s*S, H' = h+H, V' = vb+vc*V
            
			f >= 1/2
            Hi	(R-I')/S'	(G-I')/S'	(B-I')/S'
            0	1-2/3*f		-1/3+2/3*f	-2/3
            1	-1/3+2/3*f	1-2/3*f		-2/3
            2	-2/3		1-2/3*f		-1/3+2/3*f
            3	-2/3		-1/3+2/3*f	1-2/3*f
            4	-1/3+2/3*f	-2/3		1-2/3*f
            5	1-2/3*f		-2/3		-1/3+2/3*f

			f < 1/2
            Hi	(R-I')/S'	(G-I')/S'	(B-I')/S'
            0	2/3			-1/3+2/3*f	-1/3-2/3*f
            1	-1/3+2/3*f	2/3			-1/3-2/3*f
            2	-1/3-2/3*f	2/3			-1/3+2/3*f
            3	-1/3-2/3*f	-1/3+2/3*f	2/3
            4	-1/3+2/3*f	-1/3-2/3*f	2/3
            5	2/3			-1/3-2/3*f	-1/3+2/3*f
            */

			if (f>=256) {
				switch (Hi) {
				case 0:
					b = mapI[ScaledValue(3*Ia-Sa*2)];					// Ia+Sa*(-2/3)
					g = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(2*f-512)))];	// Ia+Sa*(-1/3+2/3*f)
					r = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(3*512-2*f)))];	// Ia+Sa*(1-2/3*f)
					break;
				case 1:
					b = mapI[ScaledValue(3*Ia-Sa*2)];
					g = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(3*512-2*f)))];
					r = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(2*f-512)))];
					break;
				case 2:
					b = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(2*f-512)))];
					g = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(3*512-2*f)))];
					r = mapI[ScaledValue(3*Ia-Sa*2)];
					break;
				case 3:
					b = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(3*512-2*f)))];
					g = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(2*f-512)))];
					r = mapI[ScaledValue(3*Ia-Sa*2)];
					break;
				case 4:
					b = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(3*512-2*f)))];
					g = mapI[ScaledValue(3*Ia-Sa*2)];
					r = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(2*f-512)))];
					break;
				case 5:
					b = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(2*f-512)))];
					g = mapI[ScaledValue(3*Ia-Sa*2)];
					r = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(3*512-2*f)))];
					break;
				}
			} else {
				switch (Hi) {
				case 0:
					b = mapI[ScaledValue(3*Ia-ScaledValue(Sa*(2*f+512)))];	// Ia+Sa*(-1/3-2/3*f)
					g = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(2*f-512)))];	// Ia+Sa*(-1/3+2/3*f)
					r = mapI[ScaledValue(3*Ia+Sa*2)];					// Ia+Sa*(2/3)
					break;
				case 1:
					b = mapI[ScaledValue(3*Ia-ScaledValue(Sa*(2*f+512)))];
					g = mapI[ScaledValue(3*Ia+Sa*2)];
					r = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(2*f-512)))];
					break;
				case 2:
					b = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(2*f-512)))];
					g = mapI[ScaledValue(3*Ia+Sa*2)];
					r = mapI[ScaledValue(3*Ia-ScaledValue(Sa*(2*f+512)))];
					break;
				case 3:
					b = mapI[ScaledValue(3*Ia+Sa*2)];
					g = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(2*f-512)))];
					r = mapI[ScaledValue(3*Ia-ScaledValue(Sa*(2*f+512)))];
					break;
				case 4:
					b = mapI[ScaledValue(3*Ia+Sa*2)];
					g = mapI[ScaledValue(3*Ia-ScaledValue(Sa*(2*f+512)))];
					r = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(2*f-512)))];
					break;
				case 5:
					b = mapI[ScaledValue(3*Ia+ScaledValue(Sa*(2*f-512)))];
					g = mapI[ScaledValue(3*Ia-ScaledValue(Sa*(2*f+512)))];
					r = mapI[ScaledValue(3*Ia+Sa*2)];
					break;
				}
			}
			srcp[x] = min(255,max(0,b));
			srcp[x+1] = min(255,max(0,g));
			srcp[x+2] = min(255,max(0,r));
		}
		srcp += src_pitch;
	}

	return src;
}

AVSValue __cdecl HSIAdjust::Create(AVSValue args, void* user_data, IScriptEnvironment* env)
{
    return new HSIAdjust(args[0].AsClip(),
    					 args[1].AsFloat(0.0),		// hue
    					 args[2].AsFloat(1.0),		// sat
	    				 args[3].AsFloat(0.0),		// bright
		    			 args[4].AsFloat(1.0),		// cont
    					 env);
}


extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
  env->AddFunction("HSVAdjust", "c[hue]f[sat]f[bright]f[cont]f[brightness]s", HSVAdjust::Create, 0);
  env->AddFunction("HSLAdjust", "c[hue]f[sat]f[bright]f[cont]f", HSVAdjust::CreateHSL, 0);
  env->AddFunction("HSIAdjust", "c[hue]f[sat]f[bright]f[cont]f", HSIAdjust::Create, 0);
  return 0;
}

